"""
Utility functions.
"""


def load_tf_names(path):
    """
    :param path: the path of the transcription factor list file.
    :return: a list of transcription factor names read from the file.
    """

    with open(path) as file:
        tfs_in_file = [line.strip() for line in file.readlines()]

    return tfs_in_file



"""
Core functional building blocks, composed in a Dask graph for distributed computation.
"""

import numpy as np
import pandas as pd
import logging

from sklearn.ensemble import GradientBoostingRegressor, RandomForestRegressor, ExtraTreesRegressor
from dask import delayed
from dask.dataframe import from_delayed
from dask.dataframe.utils import make_meta

logger = logging.getLogger(__name__)

DEMON_SEED = 666
ANGEL_SEED = 777
EARLY_STOP_WINDOW_LENGTH = 25

SKLEARN_REGRESSOR_FACTORY = {
    'RF': RandomForestRegressor,
    'ET': ExtraTreesRegressor,
    'GBM': GradientBoostingRegressor
}

# scikit-learn random forest regressor
RF_KWARGS = {
    'n_jobs': 1,
    'n_estimators': 1000,
    'max_features': 'sqrt'
}

# scikit-learn extra-trees regressor
ET_KWARGS = {
    'n_jobs': 1,
    'n_estimators': 1000,
    'max_features': 'sqrt'
}

# scikit-learn gradient boosting regressor
GBM_KWARGS = {
    'learning_rate': 0.01,
    'n_estimators': 500,
    'max_features': 0.1
}

# scikit-learn stochastic gradient boosting regressor
SGBM_KWARGS = {
    'learning_rate': 0.01,
    'n_estimators': 5000,  # can be arbitrarily large
    'max_features': 0.1,
    'subsample': 0.9
}


def is_sklearn_regressor(regressor_type):
    """
    :param regressor_type: string. Case insensitive.
    :return: whether the regressor type is a scikit-learn regressor, following the scikit-learn API.
    """
    return regressor_type.upper() in SKLEARN_REGRESSOR_FACTORY.keys()


def is_xgboost_regressor(regressor_type):
    """
    :param regressor_type: string. Case insensitive.
    :return: boolean indicating whether the regressor type is the xgboost regressor.
    """
    return regressor_type.upper() == 'XGB'


def is_oob_heuristic_supported(regressor_type, regressor_kwargs):
    """
    :param regressor_type: on
    :param regressor_kwargs:
    :return: whether early stopping heuristic based on out-of-bag improvement is supported.

    """
    return \
        regressor_type.upper() == 'GBM' and \
        'subsample' in regressor_kwargs and \
        regressor_kwargs['subsample'] < 1.0


def to_tf_matrix(expression_matrix,
                 gene_names,
                 tf_names):
    """
    :param expression_matrix: numpy matrix. Rows are observations and columns are genes.
    :param gene_names: a list of gene names. Each entry corresponds to the expression_matrix column with same index.
    :param tf_names: a list of transcription factor names. Should be a subset of gene_names.
    :return: tuple of:
             0: A numpy matrix representing the predictor matrix for the regressions.
             1: The gene names corresponding to the columns in the predictor matrix.
    """

    tuples = [(index, gene) for index, gene in enumerate(gene_names) if gene in tf_names]

    tf_indices = [t[0] for t in tuples]
    tf_matrix_names = [t[1] for t in tuples]

    return expression_matrix[:, tf_indices], tf_matrix_names


def fit_model(regressor_type,
              regressor_kwargs,
              tf_matrix,
              target_gene_expression,
              early_stop_window_length=EARLY_STOP_WINDOW_LENGTH,
              seed=DEMON_SEED):
    """
    :param regressor_type: string. Case insensitive.
    :param regressor_kwargs: a dictionary of key-value pairs that configures the regressor.
    :param tf_matrix: the predictor matrix (transcription factor matrix) as a numpy array.
    :param target_gene_expression: the target (y) gene expression to predict in function of the tf_matrix (X).
    :param early_stop_window_length: window length of the early stopping monitor.
    :param seed: (optional) random seed for the regressors.
    :return: a trained regression model.
    """
    regressor_type = regressor_type.upper()

    assert tf_matrix.shape[0] == len(target_gene_expression)

    def do_sklearn_regression():
        regressor = SKLEARN_REGRESSOR_FACTORY[regressor_type](random_state=seed, **regressor_kwargs)

        with_early_stopping = is_oob_heuristic_supported(regressor_type, regressor_kwargs)

        if with_early_stopping:
            regressor.fit(tf_matrix, target_gene_expression, monitor=EarlyStopMonitor(early_stop_window_length))
        else:
            regressor.fit(tf_matrix, target_gene_expression)

        return regressor

    if is_sklearn_regressor(regressor_type):
        return do_sklearn_regression()
    # elif is_xgboost_regressor(regressor_type):
    #     raise ValueError('XGB regressor not yet supported')
    else:
        raise ValueError('Unsupported regressor type: {0}'.format(regressor_type))


def to_feature_importances(regressor_type,
                           regressor_kwargs,
                           trained_regressor):
    """
    Motivation: when the out-of-bag improvement heuristic is used, we cancel the effect of normalization by dividing
    by the number of trees in the regression ensemble by multiplying again by the number of trees used.

    This enables prioritizing links that were inferred in a regression where lots of

    :param regressor_type: string. Case insensitive.
    :param regressor_kwargs: a dictionary of key-value pairs that configures the regressor.
    :param trained_regressor: the trained model from which to extract the feature importances.
    :return: the feature importances inferred from the trained model.
    """

    if is_oob_heuristic_supported(regressor_type, regressor_kwargs):
        n_estimators = len(trained_regressor.estimators_)

        denormalized_importances = trained_regressor.feature_importances_ * n_estimators

        return denormalized_importances
    else:
        return trained_regressor.feature_importances_


def to_meta_df(trained_regressor,
               target_gene_name):
    """
    :param trained_regressor: the trained model from which to extract the meta information.
    :param target_gene_name: the name of the target gene.
    :return: a Pandas DataFrame containing side information about the regression.
    """
    n_estimators = len(trained_regressor.estimators_)

    return pd.DataFrame({'target': [target_gene_name], 'n_estimators': [n_estimators]})


def to_links_df(regressor_type,
                regressor_kwargs,
                trained_regressor,
                tf_matrix_gene_names,
                target_gene_name):
    """
    :param regressor_type: string. Case insensitive.
    :param regressor_kwargs: dict of key-value pairs that configures the regressor.
    :param trained_regressor: the trained model from which to extract the feature importances.
    :param tf_matrix_gene_names: the list of names corresponding to the columns of the tf_matrix used to train the model.
    :param target_gene_name: the name of the target gene.
    :return: a Pandas DataFrame['TF', 'target', 'importance'] representing inferred regulatory links and their
             connection strength.
    """

    def pythonic():
        # feature_importances = trained_regressor.feature_importances_
        feature_importances = to_feature_importances(regressor_type, regressor_kwargs, trained_regressor)

        links_df = pd.DataFrame({'TF': tf_matrix_gene_names, 'importance': feature_importances})
        links_df['target'] = target_gene_name

        clean_links_df = links_df[links_df.importance > 0].sort_values(by='importance', ascending=False)

        return clean_links_df[['TF', 'target', 'importance']]

    if is_sklearn_regressor(regressor_type):
        return pythonic()
    elif is_xgboost_regressor(regressor_type):
        raise ValueError('XGB regressor not yet supported')
    else:
        raise ValueError('Unsupported regressor type: ' + regressor_type)


def clean(tf_matrix,
          tf_matrix_gene_names,
          target_gene_name):
    """
    :param tf_matrix: numpy array. The full transcription factor matrix.
    :param tf_matrix_gene_names: the full list of transcription factor names, corresponding to the tf_matrix columns.
    :param target_gene_name: the target gene to remove from the tf_matrix and tf_names.
    :return: a tuple of (matrix, names) equal to the specified ones minus the target_gene_name if the target happens
             to be one of the transcription factors. If not, the specified (tf_matrix, tf_names) is returned verbatim.
    """

    if target_gene_name not in tf_matrix_gene_names:
        clean_tf_matrix = tf_matrix
    else:
        clean_tf_matrix = np.delete(tf_matrix, tf_matrix_gene_names.index(target_gene_name), 1)

    clean_tf_names = [tf for tf in tf_matrix_gene_names if tf != target_gene_name]

    assert clean_tf_matrix.shape[1] == len(clean_tf_names)  # sanity check

    return clean_tf_matrix, clean_tf_names


def retry(fn, max_retries=10, warning_msg=None, fallback_result=None):
    """
    Minimalistic retry strategy to compensate for failures probably caused by a thread-safety bug in scikit-learn:
    * https://github.com/scikit-learn/scikit-learn/issues/2755
    * https://github.com/scikit-learn/scikit-learn/issues/7346

    :param fn: the function to retry.
    :param max_retries: the maximum number of retries to attempt.
    :param warning_msg: a warning message to display when an attempt fails.
    :param fallback_result: result to return when all attempts fail.
    :return: Returns the result of fn if one attempt succeeds, else return fallback_result.
    """
    nr_retries = 0

    result = fallback_result

    for attempt in range(max_retries):
        try:
            result = fn()
        except Exception as cause:
            nr_retries += 1

            msg_head = '' if warning_msg is None else repr(warning_msg) + ' '
            msg_tail = "Retry ({1}/{2}). Failure caused by {0}.".format(repr(cause), nr_retries, max_retries)

            logger.warning(msg_head + msg_tail)
        else:
            break

    return result


def infer_data(regressor_type,
               regressor_kwargs,
               tf_matrix,
               tf_matrix_gene_names,
               target_gene_name,
               target_gene_expression,
               include_meta=False,
               early_stop_window_length=EARLY_STOP_WINDOW_LENGTH,
               seed=DEMON_SEED):
    """
    Ties together regressor model training with regulatory links and meta data extraction.

    :param regressor_type: string. Case insensitive.
    :param regressor_kwargs: dict of key-value pairs that configures the regressor.
    :param tf_matrix: numpy matrix. The feature matrix X to use for the regression.
    :param tf_matrix_gene_names: list of transcription factor names corresponding to the columns of the tf_matrix used to
                                 train the regression model.
    :param target_gene_name: the name of the target gene to infer the regulatory links for.
    :param target_gene_expression: the expression profile of the target gene. Numpy array.
    :param include_meta: whether to also return the meta information DataFrame.
    :param early_stop_window_length: window length of the early stopping monitor.
    :param seed: (optional) random seed for the regressors.
    :return: if include_meta == True, return links_df, meta_df

             link_df: a Pandas DataFrame['TF', 'target', 'importance'] containing inferred regulatory links and their
             connection strength.

             meta_df: a Pandas DataFrame['target', 'meta', 'value'] containing meta information regarding the trained
             regression model.
    """
    def fn():
        (clean_tf_matrix, clean_tf_matrix_gene_names) = clean(tf_matrix, tf_matrix_gene_names, target_gene_name)

        try:
            trained_regressor = fit_model(regressor_type, regressor_kwargs, clean_tf_matrix, target_gene_expression,
                                          early_stop_window_length, seed)
        except ValueError as e:
            raise ValueError("Regression for target gene {0} failed. Cause {1}.".format(target_gene_name, repr(e)))

        links_df = to_links_df(regressor_type, regressor_kwargs, trained_regressor, clean_tf_matrix_gene_names,
                               target_gene_name)

        if include_meta:
            meta_df = to_meta_df(trained_regressor, target_gene_name)

            return links_df, meta_df
        else:
            return links_df

    fallback_result = (None, None) if include_meta else None

    return retry(fn,
                 fallback_result=fallback_result,
                 warning_msg='infer_data failed for target {0}'.format(target_gene_name))


def target_gene_indices(gene_names,
                        target_genes):
    """
    :param gene_names: list of gene names.
    :param target_genes: either int (the top n), 'all', or a collection (subset of gene_names).
    :return: the (column) indices of the target genes in the expression_matrix.
    """

    if isinstance(target_genes, list) and len(target_genes) == 0:
        return []

    if isinstance(target_genes, str) and target_genes.upper() == 'ALL':
        return list(range(len(gene_names)))

    elif isinstance(target_genes, int):
        top_n = target_genes
        assert top_n > 0

        return list(range(min(top_n, len(gene_names))))

    elif isinstance(target_genes, list):
        if not target_genes:  # target_genes is empty
            return target_genes
        elif all(isinstance(target_gene, str) for target_gene in target_genes):
            return [index for index, gene in enumerate(gene_names) if gene in target_genes]
        elif all(isinstance(target_gene, int) for target_gene in target_genes):
            return target_genes
        else:
            raise ValueError("Mixed types in target genes.")

    else:
        raise ValueError("Unable to interpret target_genes.")


def create_graph(expression_matrix,
                 gene_names,
                 tf_names,
                 regressor_type,
                 regressor_kwargs,
                 client=None,
                 target_genes='all',
                 limit=None,
                 include_meta=False,
                 early_stop_window_length=EARLY_STOP_WINDOW_LENGTH,
                 seed=DEMON_SEED):
    """
    Main API function. Create a Dask computation graph.

    :param expression_matrix: numpy matrix. Rows are observations and columns are genes.
    :param gene_names: list of gene names. Each entry corresponds to the expression_matrix column with same index.
    :param tf_names: list of transcription factor names. Should have a non-empty intersection with gene_names.
    :param regressor_type: regressor type. Case insensitive.
    :param regressor_kwargs: dict of key-value pairs that configures the regressor.
    :param client: a dask.distributed client instance.
                   * Recommended to use in a distributed environment!
                   * Used to scatter-broadcast the tf matrix to the workers instead of simply wrapping in a delayed().
                   * If None, the tf_matrix will be wrapped in a delayed(), suboptimal in a distributed setting.
    :param target_genes: either int, 'all' or a collection that is a subset of gene_names.
    :param limit: optional number of top regulatory links to return. Default None.
    :param include_meta: Also return the meta DataFrame. Default False.
    :param early_stop_window_length: window length of the early stopping monitor.
    :param seed: (optional) random seed for the regressors. Default 666.
    :return: if include_meta is False, returns a Dask graph that computes the links DataFrame.
             If include_meta is True, returns a tuple: the links DataFrame and the meta DataFrame.
    """

    assert expression_matrix.shape[1] == len(gene_names)

    tf_matrix, tf_matrix_gene_names = to_tf_matrix(expression_matrix, gene_names, tf_names)

    if client is None:
        delayed_or_future_tf_matrix = delayed(tf_matrix, pure=True)
    else:
        delayed_or_future_tf_matrix = client.scatter(tf_matrix, broadcast=True)

    delayed_tf_matrix_gene_names = delayed(tf_matrix_gene_names, pure=True)

    delayed_link_dfs = []  # collection of delayed link DataFrames
    delayed_meta_dfs = []  # collection of delayed meta DataFrame

    for target_gene_index in target_gene_indices(gene_names, target_genes):
        target_gene_name = gene_names[target_gene_index]
        target_gene_expression = expression_matrix[:, target_gene_index]

        if include_meta:
            delayed_link_df, delayed_meta_df = delayed(infer_data, pure=True, nout=2)(
                regressor_type, regressor_kwargs,
                delayed_or_future_tf_matrix, delayed_tf_matrix_gene_names,
                target_gene_name, target_gene_expression, include_meta, early_stop_window_length, seed)

            if delayed_link_df is not None:
                delayed_link_dfs.append(delayed_link_df)
                delayed_meta_dfs.append(delayed_meta_df)
        else:
            delayed_link_df = delayed(infer_data, pure=True)(
                regressor_type, regressor_kwargs,
                delayed_or_future_tf_matrix, delayed_tf_matrix_gene_names,
                target_gene_name, target_gene_expression, include_meta, early_stop_window_length, seed)

            if delayed_link_df is not None:
                delayed_link_dfs.append(delayed_link_df)

    # gather the DataFrames into one distributed DataFrame
    all_links_df = from_delayed(delayed_link_dfs,
                                meta=make_meta({'TF': str, 'target': str, 'importance': float}))
    all_meta_df = from_delayed(delayed_meta_dfs,
                               meta=make_meta({'target': str, 'n_estimators': int}))

    # optionally limit the number of resulting regulatory links, descending by top importance
    if limit:
        maybe_limited_links_df = all_links_df.nlargest(limit, columns=['importance'])
    else:
        maybe_limited_links_df = all_links_df

    # optionally return the meta DataFrame as well
    if include_meta:
        return maybe_limited_links_df, all_meta_df
    else:
        return maybe_limited_links_df


class EarlyStopMonitor:

    def __init__(self, window_length=EARLY_STOP_WINDOW_LENGTH):
        """
        :param window_length: length of the window over the out-of-bag errors.
        """

        self.window_length = window_length

    def window_boundaries(self, current_round):
        """
        :param current_round:
        :return: the low and high boundaries of the estimators window to consider.
        """

        lo = max(0, current_round - self.window_length + 1)
        hi = current_round + 1

        return lo, hi

    def __call__(self, current_round, regressor, _):
        """
        Implementation of the GradientBoostingRegressor monitor function API.

        :param current_round: the current boosting round.
        :param regressor: the regressor.
        :param _: ignored.
        :return: True if the regressor should stop early, else False.
        """

        if current_round >= self.window_length - 1:
            lo, hi = self.window_boundaries(current_round)
            return np.mean(regressor.oob_improvement_[lo: hi]) < 0
        else:
            return False



"""
Tests for arboreto.core.
"""

from random import shuffle
from os.path import join
from unittest import TestCase, skip

import dask
from distributed import Client, LocalCluster

from arboreto.core import *
from arboreto.utils import *
from tests import resources_path

net1_ex_path = join(resources_path, 'dream5/net1/net1_expression_data.tsv')
net1_tf_path = join(resources_path, 'dream5/net1/net1_transcription_factors.tsv')

net1_shape = (805, 1643)


def load_expression_matrix(path, delimiter='\t'):
    return np.genfromtxt(path, delimiter=delimiter, skip_header=1)


def load_gene_names(path, delimiter='\t'):
    with open(path) as file:
        gene_names = [gene.strip() for gene in file.readline().split(delimiter)]

    return gene_names

net1_ex_matrix = load_expression_matrix(net1_ex_path)
net1_gene_names = load_gene_names(net1_ex_path)
net1_tf_names_pure = load_tf_names(net1_tf_path)

# All tests should work with shuffled, dirty TF list.
net1_tf_names = sum([net1_tf_names_pure, ['foo', 'gee', 'bar']], [])
shuffle(net1_tf_names)

net1_tf_matrix, net1_tf_matrix_gene_names = to_tf_matrix(net1_ex_matrix,
                                                         net1_gene_names,
                                                         net1_tf_names)


class IsOobHeuristicSupportedTests(TestCase):

    def test_RF(self):
        self.assertFalse(is_oob_heuristic_supported("RF", {}))

    def test_non_stochastic_GBM(self):
        self.assertFalse(is_oob_heuristic_supported("GBM", {}))

    def test_stochastic_GBM(self):
        self.assertFalse(is_oob_heuristic_supported("GBM", {'subsample': 1}))

        self.assertTrue(is_oob_heuristic_supported("GBM", {'subsample': 0.9}))


class ToTFMatrixTests(TestCase):

    def test_TF_matrix(self):
        self.assertEquals(net1_tf_matrix.shape, (805, 195))


class InferDataTests(TestCase):

    TF = 0
    NO_TF = 200

    def inner(self, regressor_type, regressor_kwargs, target_idx, seed=DEMON_SEED):
        target_gene_name = net1_gene_names[target_idx]
        target_gene_expression = net1_ex_matrix[:, target_idx]

        links_df, meta_df = infer_data(regressor_type,
                                       regressor_kwargs,
                                       net1_tf_matrix,
                                       net1_tf_matrix_gene_names,
                                       target_gene_name,
                                       target_gene_expression,
                                       include_meta=True,
                                       seed=seed)

        self.assertEqual(meta_df['target'][0], target_gene_name)

        self.assertListEqual(list(links_df.columns), ['TF', 'target', 'importance'])

        self.assertFalse(target_gene_name in links_df['TF'].values)

    def test_smoke_fit_RF_model(self):
        self.inner("RF", RF_KWARGS, self.TF)
        self.inner("RF", RF_KWARGS, self.NO_TF)

    def test_smoke_fit_ET_model(self):
        self.inner("ET", ET_KWARGS, self.TF)
        self.inner("ET", ET_KWARGS, self.NO_TF)

    def test_smoke_fit_GBM_model(self):
        self.inner("GBM", GBM_KWARGS, self.TF)
        self.inner("GBM", GBM_KWARGS, self.NO_TF)

    def test_smoke_fit_stochastic_GBM_model(self):
        self.inner("GBM", SGBM_KWARGS, self.TF)
        self.inner("GBM", SGBM_KWARGS, self.NO_TF)

    def test_smoke_fit_stochastic_GBM_model_seed_None(self):
        self.inner("GBM", SGBM_KWARGS, self.TF, seed=None)


class RetryTests(TestCase):

    @staticmethod
    def blue_skies():
        return 1

    @staticmethod
    def i_will_never_work():
        raise ValueError('never')

    attempts = 0

    def i_procrastinate(self):
        self.attempts += 1

        if self.attempts == 3:
            return 1
        else:
            raise ValueError('later')

    def test_blue_skies(self):
        result = retry(self.blue_skies)

        self.assertEquals(result, 1)

    def test_always_fails_no_fallback(self):
        result = retry(self.i_will_never_work)

        self.assertFalse(result)

    def test_always_fails_with_fallback(self):
        result = retry(self.i_will_never_work, fallback_result=1)

        self.assertEquals(result, 1)

    def test_succeeds_after_attempts(self):
        result = retry(self.i_procrastinate)

        self.assertEquals(result, 1)


class ComputeGraphTests(TestCase):

    test_range = range(200, 205)

    @skip
    def test_net1_only_links_3_targets(self):
        graph = create_graph(net1_ex_matrix,
                             net1_gene_names,
                             net1_tf_names,
                             "GBM",
                             SGBM_KWARGS,
                             target_genes=list(self.test_range))

        network_df = graph.compute(get=dask.get)

        self.assertEquals(len(self.test_range), len(network_df['target'].unique()))

    @skip
    def test_net1_links_and_meta_3_targets(self):
        network_graph, meta_graph = create_graph(net1_ex_matrix,
                                                 net1_gene_names,
                                                 net1_tf_names,
                                                 "GBM",
                                                 SGBM_KWARGS,
                                                 target_genes=list(self.test_range),
                                                 include_meta=True,
                                                 early_stop_window_length=10)

        result = dask.compute(network_graph, meta_graph)

        network_df = result[0]
        meta_df = result[1]

        self.assertEquals(len(self.test_range), len(network_df['target'].unique()))
        self.assertEquals(len(self.test_range), len(meta_df['target'].unique()))

    def test_with_distributed_client(self):
        lc = LocalCluster(diagnostics_port=None)
        client = Client(lc)

        graph = create_graph(net1_ex_matrix,
                             net1_gene_names,
                             net1_tf_names,
                             "GBM",
                             SGBM_KWARGS,
                             target_genes=list(self.test_range),
                             client=client)

        network_df = client.compute(graph, sync=True)

        self.assertEquals(len(self.test_range), len(network_df['target'].unique()))

        client.close()
        lc.close()


class EarlyStopMonitorTests(TestCase):

    def test_window_boundaries(self):
        m = EarlyStopMonitor(window_length=10)

        self.assertEqual(m.window_boundaries(0), (0, 1))
        self.assertEqual(m.window_boundaries(1), (0, 2))
        self.assertEqual(m.window_boundaries(2), (0, 3))
        self.assertEqual(m.window_boundaries(3), (0, 4))
        self.assertEqual(m.window_boundaries(4), (0, 5))
        self.assertEqual(m.window_boundaries(5), (0, 6))
        self.assertEqual(m.window_boundaries(6), (0, 7))
        self.assertEqual(m.window_boundaries(7), (0, 8))
        self.assertEqual(m.window_boundaries(8), (0, 9))
        self.assertEqual(m.window_boundaries(9), (0, 10))
        self.assertEqual(m.window_boundaries(10), (1, 11))


class CleanTFMatrixTests(TestCase):

    # tests should run with net1_tf_names in randomized order.

    tf_matrix, tf_matrix_gene_names = to_tf_matrix(net1_ex_matrix, net1_gene_names, net1_tf_names)

    target_is_TF = "G1"
    target_not_TF = "G666"

    def test_target_is_TF(self):
        (clean_tf_matrix, clean_tf_names) = clean(self.tf_matrix, self.tf_matrix_gene_names, self.target_is_TF)

        self.assertEquals(clean_tf_matrix.shape[1], self.tf_matrix.shape[1] - 1)
        self.assertEquals(len(clean_tf_names), len(self.tf_matrix_gene_names) - 1)

        self.assertTrue(self.target_is_TF in self.tf_matrix_gene_names)
        self.assertFalse(self.target_is_TF in clean_tf_names)

    def test_target_not_TF(self):
        (clean_tf_matrix, clean_tf_names) = clean(self.tf_matrix, self.tf_matrix_gene_names, self.target_not_TF)

        self.assertEquals(clean_tf_matrix.shape, self.tf_matrix.shape)
        self.assertEquals(clean_tf_names, self.tf_matrix_gene_names)


class TargetGeneIndicesTest(TestCase):

    gene_names = ['A', 'B', 'C', 'D', 'E']

    def test_empty(self):
        self.assertEquals([], target_gene_indices(self.gene_names, []))

    def test_subset_strings(self):
        self.assertEquals([0, 2, 4], target_gene_indices(self.gene_names, ['A', 'C', 'E']))

    def test_subset_ints(self):
        self.assertEquals([0, 2, 4], target_gene_indices(self.gene_names, [0, 2, 4]))

    def test_all(self):
        self.assertEquals([0, 1, 2, 3, 4], target_gene_indices(self.gene_names, 'all'))

    def test_top(self):
        self.assertEquals([0, 1, 2], target_gene_indices(self.gene_names, 3))

        self.assertEquals([0, 1, 2, 3, 4], target_gene_indices(self.gene_names, 6))

        with self.assertRaises(AssertionError):
            target_gene_indices(self.gene_names, 0)

    def test_mixed_types(self):
        with self.assertRaises(ValueError):
            target_gene_indices(self.gene_names, ['A', 2, 3])

    def test_error(self):
        with self.assertRaises(ValueError):
            target_gene_indices(self.gene_names, 'some')


class Dream5Net1Tests(TestCase):

    def test_load_net1_matrix(self):
        self.assertEquals(net1_shape, net1_ex_matrix.shape)

    def test_load_net1_gene_names(self):
        self.assertEquals(net1_shape[1], len(net1_gene_names))

    def test_load_net1_tf_names(self):
        self.assertEquals(195, len(net1_tf_matrix_gene_names))



"""
Tests for arboreto.algo.
"""

from unittest import TestCase, skip
import numpy as np
import pandas as pd

from scipy.sparse import csc_matrix
from distributed import Client, LocalCluster
from os.path import join

from arboreto.algo import _prepare_input, _prepare_client
from arboreto.algo import grnboost2, genie3
from arboreto.utils import *
from tests import resources_path


class PrepareClientTest(TestCase):

    def test_None(self):
        client, shutdown_callback = _prepare_client(None)

        self.assertIn('127.0.0.1', client.scheduler.address)

        shutdown_callback()

    def test_local(self):
        client, shutdown_callback = _prepare_client('local')

        self.assertIn('127.0.0.1', client.scheduler.address)

        shutdown_callback()

    def test_client(self):
        lc = LocalCluster(diagnostics_port=None)
        passed = Client(lc)

        client, shutdown_callback = _prepare_client(passed)

        self.assertEquals(client, passed)

        shutdown_callback()
        lc.close()

        self.assertEquals(lc.status, 'closed')

    def test_address(self):
        with self.assertRaises(Exception) as context:
            address = 'tcp://127.0.0.2:12345'
            _prepare_client(address)

        self.assertIn('Timed out trying to connect to \'tcp://127.0.0.2:12345\'', str(context.exception))

    def test_other(self):
        with self.assertRaises(Exception) as context:
            _prepare_client(666)

        self.assertIn('Invalid client specified', str(context.exception))


zeisel_small_path = join(resources_path, 'sparse/zeisel_small.tsv')
zeisel_tfs_path = join(resources_path, 'sparse/zeisel_tfs.txt')

df = pd.read_csv(zeisel_small_path, sep='\t')
tfs = load_tf_names(zeisel_tfs_path)


class PrepareInputTest(TestCase):

    def test_DataFrame(self):
        m, g, t = _prepare_input(expression_data=df,
                                 gene_names=None,
                                 tf_names=tfs)

        self.assertTrue(isinstance(m, np.ndarray))
        self.assertEquals((500, 50), m.shape)
        self.assertEquals(50, len(g))
        self.assertEquals(4, len(t))

    def test_numpy_dense_matrix(self):
        m, g, t = _prepare_input(expression_data=df.as_matrix(),
                                 gene_names=list(df.columns),
                                 tf_names=tfs)

        self.assertTrue(isinstance(m, np.ndarray))
        self.assertEquals((500, 50), m.shape)
        self.assertEquals(50, len(g))
        self.assertEquals(4, len(t))

    def test_scipy_csc_matrix(self):
        csc = csc_matrix(df.as_matrix())

        m, g, t = _prepare_input(expression_data=csc,
                                 gene_names=list(df.columns),
                                 tf_names=tfs)

        self.assertTrue(isinstance(m, csc_matrix))
        self.assertEquals((500, 50), m.shape)
        self.assertEquals(50, len(g))
        self.assertEquals(4, len(t))


class LaunchTests(TestCase):

    def test_launch_grnboost2(self):
        network_df = grnboost2(df, tf_names=tfs)

        self.assertGreater(len(network_df), 100)

    def test_launch_genie3(self):
        network_df = genie3(df, tf_names=tfs)

        self.assertGreater(len(network_df), 100)



"""
Top-level functions.
"""

import pandas as pd
from distributed import Client, LocalCluster
from arboreto.core import create_graph, SGBM_KWARGS, RF_KWARGS, EARLY_STOP_WINDOW_LENGTH


def grnboost2(expression_data,
              gene_names=None,
              tf_names='all',
              client_or_address='local',
              early_stop_window_length=EARLY_STOP_WINDOW_LENGTH,
              limit=None,
              seed=None,
              verbose=False):
    """
    Launch arboreto with [GRNBoost2] profile.

    :param expression_data: one of:
           * a pandas DataFrame (rows=observations, columns=genes)
           * a dense 2D numpy.ndarray
           * a sparse scipy.sparse.csc_matrix
    :param gene_names: optional list of gene names (strings). Required when a (dense or sparse) matrix is passed as
                       'expression_data' instead of a DataFrame.
    :param tf_names: optional list of transcription factors. If None or 'all', the list of gene_names will be used.
    :param client_or_address: one of:
           * None or 'local': a new Client(LocalCluster()) will be used to perform the computation.
           * string address: a new Client(address) will be used to perform the computation.
           * a Client instance: the specified Client instance will be used to perform the computation.
    :param early_stop_window_length: early stop window length. Default 25.
    :param limit: optional number (int) of top regulatory links to return. Default None.
    :param seed: optional random seed for the regressors. Default None.
    :param verbose: print info.
    :return: a pandas DataFrame['TF', 'target', 'importance'] representing the inferred gene regulatory links.
    """

    return diy(expression_data=expression_data, regressor_type='GBM', regressor_kwargs=SGBM_KWARGS,
               gene_names=gene_names, tf_names=tf_names, client_or_address=client_or_address,
               early_stop_window_length=early_stop_window_length, limit=limit, seed=seed, verbose=verbose)


def genie3(expression_data,
           gene_names=None,
           tf_names='all',
           client_or_address='local',
           limit=None,
           seed=None,
           verbose=False):
    """
    Launch arboreto with [GENIE3] profile.

    :param expression_data: one of:
           * a pandas DataFrame (rows=observations, columns=genes)
           * a dense 2D numpy.ndarray
           * a sparse scipy.sparse.csc_matrix
    :param gene_names: optional list of gene names (strings). Required when a (dense or sparse) matrix is passed as
                       'expression_data' instead of a DataFrame.
    :param tf_names: optional list of transcription factors. If None or 'all', the list of gene_names will be used.
    :param client_or_address: one of:
           * None or 'local': a new Client(LocalCluster()) will be used to perform the computation.
           * string address: a new Client(address) will be used to perform the computation.
           * a Client instance: the specified Client instance will be used to perform the computation.
    :param limit: optional number (int) of top regulatory links to return. Default None.
    :param seed: optional random seed for the regressors. Default None.
    :param verbose: print info.
    :return: a pandas DataFrame['TF', 'target', 'importance'] representing the inferred gene regulatory links.
    """

    return diy(expression_data=expression_data, regressor_type='RF', regressor_kwargs=RF_KWARGS,
               gene_names=gene_names, tf_names=tf_names, client_or_address=client_or_address,
               limit=limit, seed=seed, verbose=verbose)


def diy(expression_data,
        regressor_type,
        regressor_kwargs,
        gene_names=None,
        tf_names='all',
        client_or_address='local',
        early_stop_window_length=EARLY_STOP_WINDOW_LENGTH,
        limit=None,
        seed=None,
        verbose=False):
    """
    :param expression_data: one of:
           * a pandas DataFrame (rows=observations, columns=genes)
           * a dense 2D numpy.ndarray
           * a sparse scipy.sparse.csc_matrix
    :param regressor_type: string. One of: 'RF', 'GBM', 'ET'. Case insensitive.
    :param regressor_kwargs: a dictionary of key-value pairs that configures the regressor.
    :param gene_names: optional list of gene names (strings). Required when a (dense or sparse) matrix is passed as
                       'expression_data' instead of a DataFrame.
    :param tf_names: optional list of transcription factors. If None or 'all', the list of gene_names will be used.
    :param early_stop_window_length: early stopping window length.
    :param client_or_address: one of:
           * None or 'local': a new Client(LocalCluster()) will be used to perform the computation.
           * string address: a new Client(address) will be used to perform the computation.
           * a Client instance: the specified Client instance will be used to perform the computation.
    :param limit: optional number (int) of top regulatory links to return. Default None.
    :param seed: optional random seed for the regressors. Default 666. Use None for random seed.
    :param verbose: print info.
    :return: a pandas DataFrame['TF', 'target', 'importance'] representing the inferred gene regulatory links.
    """
    if verbose:
        print('preparing dask client')

    client, shutdown_callback = _prepare_client(client_or_address)

    try:
        if verbose:
            print('parsing input')

        expression_matrix, gene_names, tf_names = _prepare_input(expression_data, gene_names, tf_names)

        if verbose:
            print('creating dask graph')

        graph = create_graph(expression_matrix,
                             gene_names,
                             tf_names,
                             client=client,
                             regressor_type=regressor_type,
                             regressor_kwargs=regressor_kwargs,
                             early_stop_window_length=early_stop_window_length,
                             limit=limit,
                             seed=seed)

        if verbose:
            print('computing dask graph')

        return client \
            .compute(graph, sync=True) \
            .sort_values(by='importance', ascending=False)

    finally:
        shutdown_callback(verbose)

        if verbose:
            print('finished')


def _prepare_client(client_or_address):
    """
    :param client_or_address: one of:
           * None
           * verbatim: 'local'
           * string address
           * a Client instance
    :return: a tuple: (Client instance, shutdown callback function).
    :raises: ValueError if no valid client input was provided.
    """

    if client_or_address is None or str(client_or_address).lower() == 'local':
        local_cluster = LocalCluster(diagnostics_port=None)
        client = Client(local_cluster)

        def close_client_and_local_cluster(verbose=False):
            if verbose:
                print('shutting down client and local cluster')

            client.close()
            local_cluster.close()

        return client, close_client_and_local_cluster

    elif isinstance(client_or_address, str) and client_or_address.lower() != 'local':
        client = Client(client_or_address)

        def close_client(verbose=False):
            if verbose:
                print('shutting down client')

            client.close()

        return client, close_client

    elif isinstance(client_or_address, Client):

        def close_dummy(verbose=False):
            if verbose:
                print('not shutting down client, client was created externally')

            return None

        return client_or_address, close_dummy

    else:
        raise ValueError("Invalid client specified {}".format(str(client_or_address)))


def _prepare_input(expression_data,
                   gene_names,
                   tf_names):
    """
    Wrangle the inputs into the correct formats.

    :param expression_data: one of:
                            * a pandas DataFrame (rows=observations, columns=genes)
                            * a dense 2D numpy.ndarray
                            * a sparse scipy.sparse.csc_matrix
    :param gene_names: optional list of gene names (strings).
                       Required when a (dense or sparse) matrix is passed as 'expression_data' instead of a DataFrame.
    :param tf_names: optional list of transcription factors. If None or 'all', the list of gene_names will be used.
    :return: a triple of:
             1. a np.ndarray or scipy.sparse.csc_matrix
             2. a list of gene name strings
             3. a list of transcription factor name strings.
    """

    if isinstance(expression_data, pd.DataFrame):
        expression_matrix = expression_data.as_matrix()
        gene_names = list(expression_data.columns)
    else:
        expression_matrix = expression_data
        assert expression_matrix.shape[1] == len(gene_names)

    if tf_names is None:
        tf_names = gene_names
    elif tf_names == 'all':
        tf_names = gene_names
    else:
        if len(tf_names) == 0:
            raise ValueError('Specified tf_names is empty')

        if not set(gene_names).intersection(set(tf_names)):
            raise ValueError('Intersection of gene_names and tf_names is empty.')

    return expression_matrix, gene_names, tf_names



from setuptools import setup, find_packages

setup(
    name='arboreto',
    packages=find_packages(),
    description='Scalable gene regulatory network inference using tree-based ensemble regressors',
    long_description=open('README.rst').read(),
    url='https://github.com/tmoerman/arboreto',
    download_url='https://github.com/tmoerman/arboreto/archive/0.1.tar.gz',
    version='0.1.4',
    license='BSD 3-Clause License',
    author='Thomas Moerman',
    author_email='thomas.moerman@gmail.com',
    install_requires=['scikit-learn', 'numpy', 'scipy', 'pandas', 'dask', 'distributed'],
    platforms=['any'],
    keywords=['gene', 'regulatory', 'network', 'inference', 'regression', 'ensemble', 'scalable', 'dask']
)
